--[[

A helper script designed to extract C function declarations from
header files and produce a file that can be used as input for
mutex.lua.

The script requires two arguments even though they may not be
used. The first argument is a Lua pattern that can be empty. The
pattern may be used to search for a specific set of function i.e. say
you want to search for all files starting with mega_:
pattern -> "^mega_"

The second required argument is for a file that will be opened if the
argument length >0. The optional file should be a concatenation of all
C Lua bindings used. Only functions found in the C "blob" will be
used, if the file is provided.

--]]

function os.getopt(args, ostr)
   local arg, place = nil, 0;
   return function ()
             if place == 0 then -- update scanning pointer
                place = 1
                if #args == 0 or args[1]:sub(1, 1) ~= '-' then place = 0; return nil end
                if #args[1] >= 2 then
                   place = place + 1
                   if args[1]:sub(2, 2) == '-' then -- found "--"
                      place = 0
                      table.remove(args, 1);
                      return nil;
                   end
                end
             end
             local optopt = args[1]:sub(place, place);
             place = place + 1;
             local oli = ostr:find(optopt);
             if optopt == ':' or oli == nil then -- unknown option
                if optopt == '-' then return nil end
                if place > #args[1] then
                   table.remove(args, 1);
                   place = 0;
                end
                return '?';
             end
             oli = oli + 1;
             if ostr:sub(oli, oli) ~= ':' then -- do not need argument
                arg = nil;
                if place > #args[1] then
                   table.remove(args, 1);
                   place = 0;
                end
             else -- need an argument
                if place <= #args[1] then  -- no white space
                   arg = args[1]:sub(place);
                else
                   table.remove(args, 1);
                   if #args == 0 then -- an option requiring argument is the last one
                      place = 0;
                      if ostr:sub(1, 1) == ':' then return ':' end
                      return '?';
                   else arg = args[1] end
                end
                table.remove(args, 1);
                place = 0;
             end
             return optopt, arg;
          end
end
local op={}
for o, a in os.getopt(arg, 'p:c:e:?') do
   op[o]=a or true
end

if op["?"] then
   arg={}
else
   local narg={}
   for k,v in pairs(arg) do
      if k > 0 then
         table.insert(narg,v)
      end
   end
   arg=narg
end

if #arg < 1 then
   print"\nUsage: lua fext.lua [-p pattern] [-c cdata-blob-file] [-e exclude-file] input-files...."
   print'Example using no pattern and no cdata: lua fext.lua *.h\n'
   os.exit(1)
end

local fmt=string.format

local function fopen(name, action)
   local f=io.open(name, action or "r")
   if not f then
      print("Cannot open ",name)
      os.exit(1)
   end
   return f
end

local function trim(s)
   return s:gsub("^%s*(.-)%s*$", "%1")
end

local exclude={}
local excnt=0
if op.e then
   local fp=fopen(op.e)
   assert(fp, "Cannot open "..op.e)
   for line in fp:lines"*l" do
      exclude[trim(line)]=true
      excnt = excnt + 1
   end
   fp:close()
end

local funcT={}

print(op.p and "Pattern: "..op.p or "No pattern!")
print("Cross check C data: ", op.c and "yes" or "no")
print("Exclude:", excnt)

local sfp = fopen("excluded.txt","w")
if op.p then sfp:write"\n---------- By pattern ------------\n" end
for _,name in ipairs(arg) do
   local f=fopen(name)
   local data = f:read"*a"
   f:close()
   data=data:gsub("/%*.-%*/","")
   for ret,func,proto in data:gmatch"([%a_][%w_%*]*%s*%*?)([%a_][%w_]*)%s*(%([^%(%)]-%))%s-;" do
      --print(ret,":",func,":",proto)
      proto=proto:gsub("%s+", " ")
      if not op.p or func:match(op.p) then
         funcT[func]={ret,proto}
      else
         sfp:write(fmt("%s\n",func))
      end
   end
end

if excnt > 0 then
   sfp:write"\n---------- By exclude file ------------\n"
   for func,v in pairs(funcT) do
      if exclude[func] then
         sfp:write(fmt("%s\n",func))
         funcT[func]=nil
      end
   end
end

if op.c then
   local nfuncT={}
   sfp:write"\n---------- By code match ------------\n"
   local f=fopen(op.c)
   assert(f, "Cannot open "..op.c)
   local cdata = f:read"*a"
   f:close()
   cdata=cdata:gsub("/%*.-%*/","")
   for func,v in pairs(funcT) do
      if cdata:find(func,1,true) then
         nfuncT[func]=v
      else
         sfp:write(fmt("%s\n",func))
      end
   end
   funcT=nfuncT
else
   print"Generating wrappers for all functions!"
end
sfp:close()

local fp=fopen("functions.txt","w")

for func,v in pairs(funcT) do
   fp:write(fmt("%s %s%s;\n",v[1],func,v[2]))
end
fp:close()

